#include "astar.h"

using namespace std;


Astar::Astar(Grid igrid, position idpos, position iapos) :
    grid(igrid), dpos(idpos)
{

}

std::vector<position> Astar::getpath()
{
    // 1st step : les points communiquent ?
    ArrayMatrix zones=Zonedef::run(grid);

    if(zones[dpos.y][dpos.x]==0 ||
            zones[apos.y][apos.x]!=zones[dpos.y][dpos.x]) // pas les mêmes zones, ou unreachable
    {
        MyError err=ZONES_DIFFERENTES;
        throw err;
    } 
    
    if(zones[apos.y][apos.x] == 0 && !unwalkable_reachable(zones[dpos.y][dpos.x], zones)) // arrivée unwalkable, aucune case autour walkable
    {
        MyError err=ZONES_DIFFERENTES;
        throw err;
    }

    // 2nd step : on est dans une même zone. A* pathfinding.
    int cx=dpos.x, cy=dpos.y;

    closedlist.push_back(grid[dpos.y][dpos.x]);

    while(cx!=apos.x && cy!=apos.y)
    {
        vector<Case*> adj=adjacentToOpen(cx, cy);

        bool end=false;
        for(unsigned i=0;i<adj.size();i++)
        {
            if(adj[i]->x == apos.x && adj[i]->y == apos.y) // Si on a la destination dans la liste ouverte
            {
                end=true;
                break;
            }
        }
        if(end)
            break;

        int nbCheck=adj.size()-1;
        for(unsigned i=0;i<adj.size();i++, nbCheck--)
        {
            for(int j=0;j<nbCheck;j++)
            {
                if(adj[j]->poids > adj[j+1]->poids)
                {
                    Case* buffer=adj[j+1];
                    adj[j+1]=adj[j];
                    adj[j]=buffer;
                }
            }
        }

        closedlist.push_back(grid[cy][cx]);
        cx=adj[0]->x;
        cy=adj[0]->y;
    }

    return definepath();
}

vector<Case*> Astar::adjacentToOpen(int x, int y)
{
    vector<Case*> toAdd;
    vector<Case*> retAdj;
    try{toAdd.push_back(grid[y-1][x]);} catch(const exception &e) {} // case hors tableau
    try{toAdd.push_back(grid[y+1][x]);} catch(const exception &e) {} // case hors tableau
    try{toAdd.push_back(grid[y][x-1]);} catch(const exception &e) {} // case hors tableau
    try{toAdd.push_back(grid[y][x+1]);} catch(const exception &e) {} // case hors tableau

    for(unsigned i=0;i<closedlist.size();i++)
    {
        for(unsigned j=0;j<toAdd.size();j++)
        {
            if(closedlist[i] == toAdd[j] || !(toAdd[j]->walkable))
            {
                toAdd.erase(toAdd.begin()+j); // in closed list || unwalkable=> OSEF.
            }
        }
    }

    for(unsigned i=0;i<openlist.size();i++)
    {
        for(unsigned j=0;j<toAdd.size();j++)
        {
            if(openlist[i] == toAdd[j])
            {
                int newcost=calculePoids(toAdd[j]->x, toAdd[j]->y, grid[y][x]);
                if(newcost < toAdd[j]->poids)
                {
                    toAdd[j]->poids=newcost;
                    toAdd[j]->parent=grid[y][x];
                }
                retAdj.push_back(toAdd[j]);
                toAdd.erase(toAdd.begin()+j);
            }
        }
    }


    // tri par insertion selon le poids (poids plumes devant)
    for(unsigned i=0;i<toAdd.size();i++)
    {
        toAdd[i]->parent=grid[y][x];
        toAdd[i]->poids=calculePoids(toAdd[i]->x, toAdd[i]->y, toAdd[i]->parent);
        retAdj.push_back(toAdd[i]);
        for(unsigned j=0;j<openlist.size();j++)
        {
            if(openlist[j]->poids > toAdd[i]->poids || j == openlist.size()-1) // on est à la place de l'élem
            {
                openlist.insert(openlist.begin()+j, toAdd[i]);
                toAdd.erase(toAdd.begin()+i);
            }
        }
    }

    return retAdj;
}

int Astar::calculePoids(int x, int y, Case* forceParent)
{
    if(forceParent==NULL)
        forceParent=grid[y][x]->parent;
    int fromStart=0, heuristic=0;


    // fromstart
    Case* node=grid[y][x];
    while(node!=grid[dpos.y][dpos.x])
    {
        fromStart++;
        node=node->parent;
    }

    //heuristic
    heuristic+=abs(x-apos.x);
    heuristic+=abs(y-apos.y);

    return heuristic+fromStart;
}

vector<position> Astar::definepath()
{
    vector<Case*> path;
    vector<position> pathinvert;

    Case* curr=closedlist[closedlist.size()-1];
    while(curr->parent!=NULL)
    {
        path.push_back(curr);
        curr=curr->parent;
    }

    // bourrinage pur. Mais ça marche.
    for(unsigned i=path.size()-1;i>=0;i--)
    {
        position d;
        d.x=path[i]->x;
        d.y=path[i]->y;
        pathinvert.push_back(d);
    }

    return pathinvert;
}

bool Astar::unwalkable_reachable(short area, ArrayMatrix zones)
{
    try{
        if(zones[apos.y-1][apos.x] == area)
            return true;
    } catch (const exception& e) {} // on ignore l'exception
    try{
        if(zones[apos.y+1][apos.x] == area)
            return true;
    } catch (const exception& e) {} // on ignore l'exception
    try{
        if(zones[apos.y][apos.x-1] == area)
            return true;
    } catch (const exception& e) {} // on ignore l'exception
    try{
        if(zones[apos.y][apos.x+1] == area)
            return true;
    } catch (const exception& e) {} // on ignore l'exception
    
    return false;
}


